"""Unit tests for module dynamic_delay_dask"""

import numpy as np
import pytest

from lisainstrument.dynamic_delay_dask import (
    make_dynamic_shift_lagrange_dask,
    make_dynamic_shift_linear_dask,
    numpyfy_dask_multi,
)
from lisainstrument.dynamic_delay_numpy import (
    ShiftBC,
    make_dynamic_shift_lagrange_numpy,
)
from lisainstrument.regular_interpolators import make_lagrange_polynomials


def test_dynamic_shift_linear_dask() -> None:
    t, dt = np.linspace(-5.345, 10.345, 103, retstep=True)

    def g(x):
        return 4.32546 + 3.34324 * x

    y = g(t)

    d = (0.93456456 + 0.0235345 * np.cos(4.3354 * t)) / dt

    op_da = make_dynamic_shift_linear_dask(
        d.min(), d.max(), ShiftBC.FLAT, ShiftBC.EXCEPTION
    )
    op_np = numpyfy_dask_multi(op_da, chunks=19)

    s_da = op_np(y, -d)
    s_ex = g(np.maximum(t[0], t - d * dt))

    assert s_ex == pytest.approx(s_da, abs=1e-15, rel=1e-14)


def test_dynamic_shift_lagrange_dask() -> None:
    order = 5

    t, dt = np.linspace(-5.345, 10.345, 1003, retstep=True)

    def g(x):
        return (
            4.32546
            + 3.34324 * x
            + 4.342 * x**2
            + 0.46 * x**3
            + 1.43598 * x**4
            + 3.3456 * x**5
        )

    y = g(t)

    d = (0.93456456 + 0.0235345 * np.cos(4.3354 * t)) / dt

    op_da = make_dynamic_shift_lagrange_dask(
        order, d.min(), d.max(), ShiftBC.FLAT, ShiftBC.EXCEPTION
    )
    op_na = numpyfy_dask_multi(op_da, chunks=19)

    s_da: np.ndarray = op_na(y, -d)
    s_ex = g(np.maximum(t[0], t - d * dt))

    assert s_ex[op_da.margin_left :] == pytest.approx(
        s_da[op_da.margin_left :], abs=1e-15, rel=1e-13
    )

    op_np = make_dynamic_shift_lagrange_numpy(
        order, d.min(), d.max(), ShiftBC.FLAT, ShiftBC.EXCEPTION
    )

    s_np = op_np(y, -d)

    assert np.all(s_np == s_da)


def test_lagrange_polynomials():
    length = 11
    d = -(length // 2)
    lagps = make_lagrange_polynomials(length, d)
    for j, p in enumerate(lagps):
        for x in range(d, d + length):
            if x - d == j:
                assert p(x) == pytest.approx(1, abs=0, rel=1e-12)
            else:
                assert p(x) == pytest.approx(0, abs=1e-11, rel=0)
